# Generated by Django 2.2.10 on 2020-05-12 10:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("storage", "0004_calculate_hashes"),
    ]

    operations = [
        migrations.AddField(
            model_name="referencedpath",
            name="storage_locations",
            field=models.ManyToManyField(
                related_name="files", to="storage.StorageLocation"
            ),
        ),
    ]
