# Generated by Django 2.2.10 on 2020-08-21 05:38

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0045_unreferenced_storages"),
        ("storage", "0007_remove_referencedpath_file_storage"),
    ]

    operations = [
        migrations.AddField(
            model_name="accesslog",
            name="cause",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="access_logs",
                to="flow.Data",
            ),
        ),
    ]
