# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-03-26 08:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0013_migration_history"),
    ]

    operations = [
        migrations.AddField(
            model_name="data",
            name="process_cores",
            field=models.PositiveSmallIntegerField(default=0),
        ),
        migrations.AddField(
            model_name="data",
            name="process_memory",
            field=models.PositiveIntegerField(default=0),
        ),
    ]
